; File Name   :	forth_6k.bin
; JU+TE-FORTH (FORTH 83) fr TINY-Computer, s. JU+TE 01/1990, S. 70-73
; reass: V. Pohlers, 06/2008

; Voraussetzung ist ES 4K, 8 KByte RAM (Platz 1 voll bestckt)
; Start auf %E000
; ab %FC00 Magnetbandpuffer, bis dahin frei fr Prg. (ca. 2KByte)
; Speichern von %E000 bis DP, am einfachsten bis %FBFF
; 173 Systemworte

; forth_4k.jtc.bin entspricht FORTH_NEU.BIN
; forth_6k.jtc.bin ist ein gepatchtes FORTH_NEU.BIN

;ACHTUNG: Der Code ist reassmbliert! Im Code enthaltene Konstanten
;knnten bei Verschieben der Codebereichs oder ndern des Codes Probleme
;verursachen! Wenn jemand vorhat, dieses FORTH weiter zu entwickeln,
;sollte es erst vollstndig kommentiert sein. Ein Text-Vergleich mit der 
;4K-Version ist ebenfalls hilfreich!

		cpu	z8601

;AS-Funktionen
hi              function x,(x>>8)&255
lo              function x, x&255


; Segment type:	Internal processor memory & SFR

p0	equ	0				; Port 0
p1	equ	1				; Port 1
p2	equ	2				; Port 2
p3	equ	3				; Port 3

;sio	equ	0F0h				; Serial I/O
;tmr	equ	0F1h				; Timer	mode
;t1	equ	0F2h				; Timer/counter	1
;pre1	equ	0F3h				; T1 prescaler
;t0	equ	0F4h				; Timer/counter	0
;pre0	equ	0F5h				; T0 prescaler
;p2m	equ	0F6h				; Port 2 mode
;p3m	equ	0F7h				; Port 3 mode
;p01m	equ	0F8h				; Ports	0-1 mode
;ipr	equ	0F9h				; Interrupt priority register
;irq	equ	0FAh				; Interrupt request register
;imr	equ	0FBh				; Interrupt mask register
;flags	equ	0FCh				; Program control flags
;rp	equ	0FDh				; Register pointer
sph	equ	0FEh				; Stack pointer
spl	equ	0FFh		                ; Stack pointer

; TINY-BASIC-Aufrufe
; 0BAh		p_mult
; 0E0h		p_div
; 11Fh		p_mod
; 124h		? zu mod

; ES-4.0-Aufrufe
CHARIN		equ	815h		; Zeicheneingabe
CHAROUT		equ	818h		; Zeichenausgabe
CKEY		equ	81Bh		; momentan gedrckte Taste
ES_DATA		equ	82Ah		; ZBS++
TSK		equ	17F4h		; Tastenabfrage, WKEY nach %5A
;F700 Eingabepuffer CHARIN (80 Zeichen)
;FC00 BWS

;VAR
TIBBUF		EQU	0F300h		; F300 ff. Eingabepuffer
RPSTACK		EQU	0F700h		; F700-1 down RP Stack
;stack F700h ?
;CSP	0F6FCh
;0FFFFh (DPL)

;
; Aufbau
;
; ----------------------
; |Countbyte 100XXXXX  |
; ----------------------
; |Name ASCII          |
; ----------------------
; |Linkfeld (High, Low)|
; ----------------------
; |Kennbyte            |
; ----------------------
; |Maschinencode       |
; |(Rumpf, = CFA = PFA)|
; ----------------------
;
; Kennbyte Kennzeichnung immediate und zur Steuerung
; von Strukturwrtern wie DO und zur Steuerung des Rckbersetzens
;  normal			00
;  immediate			01
;  VAR, CON			0A
;  DLIT				0C
;  DOTQ				0E
;  THENN			10
;  AGAIN, LOOP, PLOOP, UNTIL	22
;  REPEAT			32
;  IFF, WHILE			42
;  ELSEE			52
;  DO, BEGIN			80
;
; RP     Register %10
; Stack: r2,r3  TOS
;	 Register %20..%4F
;        r4..r11 Arbeitsregister
; SP     r15
; Returnstack ist Z8-Stack SPH/SPL
; ===========================================================================


 		org 	0D000h

;------------------------------------------------------------------------------
; Start JU+TE-FORTH
; ber BASIC oder ber PROG mglich
;------------------------------------------------------------------------------

START:		jp	start2
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
	
;------------------------------------------------------------------------------
; Unterprogramme und Hilfsroutinen
;------------------------------------------------------------------------------

; zu MSTAR
mmult:		call	DROP
		ld	R8, R2
		xor	R8, R4
		or	R2, R2
		jr	PL, mmult1
		com	R2
		com	R3
		incw	RR2
mmult1:		or	R4, R4
		jr	PL, mmult2
		com	R4
		com	R5
		incw	RR4
mmult2:		ld	R9, #10h
		ld	R10, #0
		ld	R11, #0
mmult3:		rrc	R4
		rrc	R5
		jr	NC, mmult4
		add	R11, R3
		adc	R10, R2
mmult4:		rrc	R10
		rrc	R11
		rrc	R6
		rrc	R7
		djnz	R9, mmult3
		ld	R2, R10
		ld	R3, R11
		ret

; zu MSLAS
mdiv1:		call	DROP
		ld	R8, R2
		xor	R8, R4
		or	R4, R4
		jr	PL, mdiv2
		com	R4
		com	R5
		incw	RR4
mdiv2:		ld	R10, R4
		ld	R11, R5
		call	DROP
		or	R4, R4
		jr	PL, mdiv3
		com	R4
		com	R5
		com	R2
		com	R3
		incw	RR2
		jr	NZ, mdiv3
		incw	RR4
mdiv3:		clr	R6
		clr	R7
		ld	R9, #20h
		rlc	R3
		rlc	R2
		rlc	R5
		rlc	R4
mdiv4:		rlc	R7
		rlc	R6
		sub	R7, R11
		sbc	R6, R10
		jr	NC, mdiv5
		add	R7, R11
		adc	R6, R10
mdiv5:		ccf
		rlc	R3
		rlc	R2
		rlc	R5
		rlc	R4
		djnz	R9, mdiv4
		ret

; TOSTACK RR4 auf Stack ablegen (PUSH)
TOSTACK:	call	sub_E0A7	; dec sp+Test auf Stackunterlauf
		ld	@R15, R2	; TOP->Stack
		dec	R15
		ld	@R15, R3
		ld	R2, R4		; RR4->TOS
		ld	R3, R5
		ret

; Test auf Stackunterlauf (DUP, TOSTACK)
sub_E0A7:	dec	R15		; SP
		ld	R6, #'7'
		cp	R15, #21h	; SP lower bound overflow?
		jp	C, error	; Ausgabe Fehler #MSG <R6>
		ret

error:		call	DOTQ		; Ausgabe Fehler #MSG <R6>
		db	 6
		db	0Dh,"#MSG "
		nop
		ld	5Ah, R6
		call	ZBS	; Ausgabe 5Ah
		nop
		nop
		nop
		jr	start3

;------------------------------------------------------------------------------
; Fortsetung Start und Hauptschleife
;------------------------------------------------------------------------------

start2:		srp	#10h		; RP
		call	DOTQ        
                db	16,0Ch,"J+T-F83 ES4.0 ",0Dh
start3:         ld      R15, #50h	; SP:=50h
                di
                srp     #0F0h		; system register
                ld      R15, #lo(RPSTACK)	; SPL
                ld      R14, #hi(RPSTACK)	; SPH	RP:=0f700h
                jr      loc_D0FC

;Patch statt ES1988-ZBS, Ausgabe 5Ah
ZBS:       	push    15h
                ld      15h, 5Ah
                call    CHAROUT	; ES Zeichenausgabe
                pop     15h
                ret
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
;
loc_D0FC:       ei
; uerer Interpreter		
start4:		srp	#10h
		call	DOTQ
		db 	4," OK",0Dh
		call	QUERY
		call	INTER
		jp	start4

;------------------------------------------------------------------------------
; Hilfsroutinen
;------------------------------------------------------------------------------

; zu HOLD
sub_E110:	ld	R2, R4		; TOS = NOS
		ld	R3, R5
		ret

; zu query1
sub_E115:	ld	R8, #0
		lde	@RR4, R8
		ret

; zu DLIT
sub_E11A:	call	TOSTACK
		ld	R10, #14h	; ?
		ret

; zu MIN, MAX, USLASH
sub_E120:	call	DROP
		ld	R6, R2
		ld	R7, R3
		sub	R7, R5
		sbc	R6, R4
		ret

; zu DPLUS, sub_E157
sub_E12C:	call	DROP
		ld	R8, R2
		ld	R9, R3
		ld	R10, R4
		ld	R11, R5
		call	DROP
		jp	DROP

; zu inter1
sub_E13D:	call	HERE
		call	TWOP
		ld	R12, R2
		ld	R13, R3
		jp	DROP

; zu inter1
sub_E14A:	call	STATE
		jp	AT

; zu NEGAT
loc_E150:	com	R2
		com	R3
		incw	RR2
		ret

; zu DMINUS, DEQU, sub_E163
sub_E157:	call	sub_E12C
		sub	R3, R9
		sbc	R2, R8
		sbc	R5, R11
		sbc	R4, R10
		ret

; zu DLESS, DULESS
sub_E163:	call	sub_E157
		clr	R2		; TOS = 0
		clr	R3
		ret

; zu SEMIS
sub_E16B:	call	STATE
		call	AT
		ld	R4, R2
		or	R4, R3
		ld	R6, #'D'
		jp	Z, error	; Ausgabe Fehler #MSG <R6>
		jp	DROP

; zu view4 (hier nicht genutzt !!)
sub_E17D:	clr	6Dh
		call	0C56h		; ES1988: TAS, dynamische Tastenabfrage mit %6D: ASCII
		jr	NZ, sub_E17D
		clr	6Dh
		jp	TOSTACK

; zu DOT, UDOT
sub_E189:	call	TYPE
		jp	SPACE

; zu view12 (hier nicht genutzt !!)
sub_E18F:	ld	5Ah, #22h	; '"'
		call	ZBS	; Ausgabe %5A
		jp	SPACE

;------------------------------------------------------------------------------
; Speicher fr Zeichen-Konvertierung
;------------------------------------------------------------------------------
; vp: in Originalcode steht "....6913579691-20121-F412" in Buffer
		
		align 100h
MDIGS:		equ	$		; xx00-Adr Absicht!
					; in EDIGS wird die Lnge per COM berechnet

;------------------------------------------------------------------------------
; DROP (16b -- )
; Entfernen von	16b aus	dem Stack
;------------------------------------------------------------------------------

NFA_DROP:	db	84h
		db	"DROP"
		dw	0
		db	0
;
DROP:		ld	R4, R2		; RR4=TOS (wird in Callern genutzt)
		ld	R5, R3
		ld	R3, @R15	; TOS=Stack
		inc	R15		; pop stack
		ld	R2, @R15
drop1:		inc	R15
		ld	R6, #'1'
		cp	R15, #51h	; Stack underflow?
		jp	NC, error	; Ausgabe Fehler #MSG <R6>
		ret

;------------------------------------------------------------------------------
; ." (--)
; Ausgabe aller	folgenden Zeichen ausschlielich des abschlieenden "
; und des ersten nach ." stehenden Leerzeichens auf dem Bildschirm.
;------------------------------------------------------------------------------

NFA_DOTQ:	db	82h
		db	'.',22h
		dw	NFA_DROP
		db	0Eh
;
DOTQ:		pop	R4
		pop	R5
		ldc	R8, @RR4
		incw	RR4
dotq1:		ld	R7, #5Ah
		ldci	@R7, @RR4
		call	ZBS        ; Ausgabe %5A
		djnz	R8, dotq1
		jp	@RR4

;------------------------------------------------------------------------------
; QUERY
; JU+TE-FORTH f. INPUT, anstelle SPAN+CONVERT
;------------------------------------------------------------------------------

NFA_QUERY:	db	85h
		db	"QUERY"
		dw	NFA_DOTQ
		db	0
;patch
QUERY:		jp	query0
	
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh

; CLS (patch)
CLS:		ld      5Ah, #0Ch
		jp      ZBS        ; Ausgabe %5A

;------------------------------------------------------------------------------
; TIB (-- adr)
; adr ist die Adresse des Text-Eingabe-Puffers mit wenigstens 80 byte Lnge.
;------------------------------------------------------------------------------

NFA_TIB:	db	83h
		db	"TIB"
		dw	NFA_QUERY
		db	0
;
TIB:		call	VAR
		dw	TIBBUF+000FFh

;------------------------------------------------------------------------------
; ! (16b adr -- )
; 16b wird in den Speicherzellen adr und adr+1 gespeichert
;------------------------------------------------------------------------------

NFA_STORE:	db	81h
		db	'!'
		dw	NFA_TIB
		db	0
;
STORE:		call	DROP		; RR4=TOS=adr
		ldc	@RR4, R2	; RR2=TOS=16b
		incw	RR4
		ldc	@RR4, R3
		jp	DROP		; 16b entfernen

;------------------------------------------------------------------------------
; VAR
;------------------------------------------------------------------------------

NFA_VAR:	db	83h
		db	"VAR"
		dw	NFA_STORE
		db	0Ah
;
VAR:		call	TOSTACK		; RR4 auf Stack
		pop	R2		; TOS = Adr.
		pop	R3
		ret

;------------------------------------------------------------------------------
; CON
;------------------------------------------------------------------------------

NFA_CON:	db	83h
		db	"CON"
		dw	NFA_VAR
		db	0Ah
;
CON:		call	TOSTACK
		pop	R6
		pop	R7
		ld	R10, #12h	; 12h = RR2 = TOS
		ldci	@R10, @RR6
		ldci	@R10, @RR6
		ret

;------------------------------------------------------------------------------
; LIT
;------------------------------------------------------------------------------

NFA_LIT:	db	83h
		db	"LIT"
		dw	NFA_CON
		db	0Ah
;
LIT:		pop	R8
		pop	R9
		ld	R10, #14h	; 14h = RR4
lit1:		ldci	@R10, @RR8
		ldci	@R10, @RR8
		call	TOSTACK		; RR4 auf Stack
		jp	@RR8

;------------------------------------------------------------------------------
; DLIT
;------------------------------------------------------------------------------

NFA_DLIT:	db	84h
		db	"DLIT"
		dw	NFA_LIT
		db	0Ch
;
DLIT:		pop	R8
		pop	R9
		call	sub_E11A
		ldci	@R10, @RR8
		ldci	@R10, @RR8
		ld	R10, #12h	; 12h = RR2 = TOS
		jr	lit1

;------------------------------------------------------------------------------
; ROT (16b1 16b2 16b3 --16b2 16b3 16b1)
; Rotieren der drei obersten Stapeleintrge,
; so da der drittoberste an die Spitze gelangt.
;------------------------------------------------------------------------------

NFA_ROT:	db	83h
		db	"ROT"
		dw	NFA_DLIT
		db	0
;
ROT:		ld	R5, @R15
		ld	@R15, R3
		inc	R15
		ld	R4, @R15
		ld	@R15, R2
		inc	R15
		ld	R3, @R15
		ld	@R15, R5
		call	drop1
		ld	R2, @R15
		ld	@R15, R4
		sub	R15, #3
		ret

;------------------------------------------------------------------------------
; SWAP (16b1 16b2 -- 16b2 16b1)
; Vertauschen der beiden obersten Stapeleintrge.
;------------------------------------------------------------------------------

NFA_SWAP:	db	84h
		db	"SWAP"
		dw	NFA_ROT
		db	0
;
SWAP:		ld	R5, @R15	; Stack->temp 
		ld	@R15, R3	; TOS->Stack
		call	drop1
		ld	R4, @R15
		ld	@R15, R2
		ld	R2, R4		; temp->TOS
		ld	R3, R5
		dec	R15
		ret

;------------------------------------------------------------------------------
; DUP (16b --16b 16b)
; Duplizieren des TOS
;------------------------------------------------------------------------------

NFA_DUP:	db	83h
		db	"DUP"
		dw	NFA_SWAP
		db	0
;
DUP:		call	sub_E0A7
		ld	@R15, R2	; TOS -> Stack
		dec	R15
		ld	@R15, R3
		ret

;------------------------------------------------------------------------------
; ?DUP (16b -- 16b 16b) oder (0 --0) .
; Duplizieren des TOS, wenn er ungleich Null ist.
;------------------------------------------------------------------------------

NFA_QDUP:	db	84h
		db	"?DUP"
		dw	NFA_DUP
		db	0
;
QDUP:		ld	R4, R2
		or	R4, R3
		jr	NZ, DUP
		ret

;------------------------------------------------------------------------------
; CLS
;------------------------------------------------------------------------------

NFA_CLS:	db	83h
		db	"CLS"
		dw	NFA_QDUP
		db	0
;
		jp	CLS	; cls

;------------------------------------------------------------------------------
; OVER (16b1 16b2 -- 16b1 16b2 16b1)
; Duplizieren des zweitobersten Stapeleintrags.
;------------------------------------------------------------------------------

NFA_OVER:	db	84h
		db	"OVER"
		dw	NFA_CLS
		db	0
;
OVER:		ld	R5, @R15	; RR4 = SEC
		inc	R15
		ld	R4, @R15
		dec	R15
		jp	TOSTACK		; RR4 -> TOP

;------------------------------------------------------------------------------
; >R (16b -- )
; Transport des TOS zum Return-Stack.
;------------------------------------------------------------------------------

NFA_TOR:	db	82h
		db	">R"
		dw	NFA_OVER
		db	0
;
TOR:		pop	R4		; RET-Wert sichern
		pop	R5
		push	R3		; TOS -> RS
		push	R2
		ld	R3, @R15	; Dec Stack (drop)
		inc	R15
		ld	R2, @R15
		call	drop1
		jp	@RR4		; RET-Adr.

;------------------------------------------------------------------------------
; R> (-- 16b)
; Transport des obersten Return-Stack-Eintrags zum Datenstack.
;------------------------------------------------------------------------------

NFA_RFROM:	db	82h
		db	"R>"
		dw	NFA_TOR
		db	0
;
RFROM:		pop	R8		; RET-Wert sichern
		pop	R9
		call	TOSTACK		; RR4 auf Stack
		pop	R2		; TOS = RR4
		pop	R3
		jp	@RR8		; RET-Adr.

;------------------------------------------------------------------------------
; R
;------------------------------------------------------------------------------

NFA_RR:		db	81h
		db	'R'
		dw	NFA_RFROM
		db	0
;
RR:		call	TOSTACK
		pop	R4
		pop	R5
		pop	R2
		pop	R3
		push	R3
		push	R2
		jp	@RR4

;------------------------------------------------------------------------------
; + (w1 w2 -- w3)
; w3 ist die Summe von wl plus w2.
;------------------------------------------------------------------------------

NFA_PLUS:	db	81h
		db	'+'
		dw	NFA_RR
		db	0
;
PLUS:		add	R3, @R15
		inc	R15
		adc	R2, @R15
		jp	drop1

;------------------------------------------------------------------------------
; - (w1 w2 -- w3)
; w3 ist die Differenz w1 minus w2.
;------------------------------------------------------------------------------

NFA_MINUS:	db	81h
		db	'-'
		dw	NFA_PLUS
		db	0
;
MINUS:		nop
		nop
		call	DROP
		sub	R3, R5
		sbc	R2, R4
		ret

;------------------------------------------------------------------------------
; * (n1 n2 -- n3)
; n3 ist das Produkt von n1 und n2.
;------------------------------------------------------------------------------

NFA_STAR:	db	81h
		db	'*'
		dw	NFA_MINUS
		db	0
;
STAR:		call	DROP
		jp	0BAh		; in TINY-BASIC-ROM

;------------------------------------------------------------------------------
; / (n1 n2 -- n3)
; n3 ist der Quotient n1/n2.
;------------------------------------------------------------------------------

NFA_SLASH:	db	81h
		db	'/'
		dw	NFA_STAR
		db	0
;
SLASH:		call	DROP
		jp	0E0h		; in TINY-BASIC-ROM

;------------------------------------------------------------------------------
; MOD (n1 n2 -- n3)
; n3 ist der Divisionsrest aus n1/n2.
;------------------------------------------------------------------------------

NFA_MOD:	db	83h
		db	"MOD"
		dw	NFA_SLASH
		db	0
;
MOD:		call	DROP
		jp	11Fh		; in TINY-BASIC-ROM

;------------------------------------------------------------------------------
; 1+ (w1 -- w2)
; w2 ist die Summe w1 plus 1.
;------------------------------------------------------------------------------

NFA_ONEP:	db	82h
		db	"1+"
		dw	NFA_MOD
		db	0
;
ONEP:		incw	RR2
		ret

;------------------------------------------------------------------------------
; NEGATE (n1 -- n2)
; n2 ist die Differenz 0 minus n1.
;------------------------------------------------------------------------------

NFA_NEGAT:	db	86h
		db	"NEGATE"
		dw	NFA_ONEP
		db	0
;
NEGAT:		jp	loc_E150

;------------------------------------------------------------------------------
; ABS (n -- u)
; u ist der Absolutbetrag von n.
;------------------------------------------------------------------------------

		nop			; ???

NFA_ABS:	db	83h
		db	"ABS"
		dw	NFA_NEGAT
		db	0
;
ABS:		or	R2, R2
		jr	MI, NEGAT
		ret

;------------------------------------------------------------------------------
; U*
;------------------------------------------------------------------------------

NFA_USTAR:	db	82h
		db	"U*"
		dw	NFA_ABS
		db	0
;
USTAR:		call	DROP
		call	mmult2
		jr	mstar1

;------------------------------------------------------------------------------
; M*
;------------------------------------------------------------------------------

		db	 0			; ???
		db	0EFh 			; ???
		db	0F3h 			; ???
		db	0F7h 			; ???
		db	0AFh 			; ???

NFA_MSTAR:	db	82h
		db	"M*"
		dw	NFA_USTAR
		db	0
;
MSTAR:		call	mmult
		or	R8, R8
		jr	PL, mstar1
		com	R2
		com	R3
		com	R6
		com	R7
		incw	RR6
		jr	NZ, mstar1
		incw	RR2
mstar1:		dec	R15
		ld	@R15, R6
		dec	R15
		ld	@R15, R7
		ret

;------------------------------------------------------------------------------
; M/
;------------------------------------------------------------------------------

NFA_MSLAS:	db	82h
		db	"M/"
		dw	NFA_MSTAR
		db	0
;
MSLAS:		call	mdiv1
		or	R8, R8
		jr	PL, mstar1
		com	R2
		com	R3
		incw	RR2
		com	R6
		com	R7
		incw	RR6
		jr	mstar1

;------------------------------------------------------------------------------
; U/
;------------------------------------------------------------------------------

NFA_USLAS:	db	82h
		db	"U/"
		dw	NFA_MSLAS
		db	0
;
USLAS:		call	DROP
		ld	R10, R4
		ld	R11, R5
		call	DROP
		call	mdiv3
		jr	mstar1

;------------------------------------------------------------------------------
; */MOD
;------------------------------------------------------------------------------

NFA_SSMOD:	db	85h
		db	"*/MOD"
		dw	NFA_USLAS
		db	0
;
SSMOD:		call	ROT
		call	ROT
		call	MSTAR
		call	ROT
		jr	MSLAS

;------------------------------------------------------------------------------
; */ (n1 n2 n3 -- n4)
; n1 und n2 werden zu einem doppelt genauen Zwischenergebnis multipliziert,
; das nach Division durch n3 dann n4 ergibt.
;------------------------------------------------------------------------------

NFA_SSLA:	db	82h
		db	"*/"
		dw	NFA_SSMOD
		db	0
;
SSLA:		call	SSMOD
		call	SWAP
		jp	DROP

;------------------------------------------------------------------------------
; */MOD (n1 n2 n3 -- n4 n5)
; Berechnung von n4 wie mit */, jedoch zustzliche bergabe
; des Divisionsrestes n5.
;------------------------------------------------------------------------------

NFA_MSMOD:	db	85h
		db	"M/MOD"
		dw	NFA_SSLA
		db	0
;
MSMOD:		call	DROP
		ld	R10, R4
		ld	R11, R5
		call	DROP
		call	mdiv3
		dec	R15
		ld	@R15, R6
		dec	R15
		ld	@R15, R7
		jp	TOSTACK

;------------------------------------------------------------------------------
; DO (w1 w2 -- )
; Erffnung einer Schleife mit Indexlauf von Anfangswert w2 bis Endwert w1.
;------------------------------------------------------------------------------

NFA_DO:		db	82h
		db	"DO"
		dw	NFA_MSMOD
		db	80h
;
DO:		call	DROP
		pop	R8
		pop	R9
		push	R3
		push	R2
		push	R5
		push	R4
		call	DROP
		jp	@RR8

;------------------------------------------------------------------------------
; +LOOP (n -- )
; Addieren von n zum Schleifenindex und Verlassen der Schleife
; mit Beseitigen der Parameter vom Return-Stack, wenn Endwert erreicht.
; Sonst Fortsetzung hinter DO.
;------------------------------------------------------------------------------

NFA_PLOOP:	db	85h
		db	"+LOOP"
		dw	NFA_DO
		db	22h
;
PLOOP:		pop	R8
		pop	R9
		pop	R4
		pop	R5
		pop	R6
		pop	R7
		push	R7
		push	R6
		add	R5, R3
		adc	R4, R2
		push	R5
		push	R4
		ld	R3, #1Ah  ; 1Ah = RR10
		ldci	@R3, @RR8
		ldci	@R3, @RR8
		sub	R5, R7
		sbc	R4, R6
		xor	R2, R4
		ld	R7, R4
		or	R7, R5
		call	DROP
		or	R7, R7
		jr	Z, ploop1
		rl	R4
		jr	NC, ploop1
		jp	@RR10
ploop1:		pop	R4
		pop	R4
		pop	R4
		pop	R4
		jp	@RR8

;------------------------------------------------------------------------------
; LOOP ( -- )
; Inkrementieren des Schleifenindex und Verlassen der Schleife
; mit Beseitigen der Parameter vom Return-Stack, wenn Endwert erreicht.
; Sonst Fortsetzung hinter DO.
;------------------------------------------------------------------------------

NFA_LOOP:	db	84h
		db	"LOOP"
		dw	NFA_PLOOP
		db	22h
;
LOOP:		call	TOSTACK
		ld	R2, #0
		ld	R3, #1
		jr	PLOOP

;------------------------------------------------------------------------------
; I (-- w)
; w ist die Kopie des Schleifenindex.
;------------------------------------------------------------------------------

NFA_IDO:	db	81h
		db	'I'
		dw	NFA_LOOP
		db	0
;
IDO:		jp	RR

;------------------------------------------------------------------------------
; LEAVE ( -- )
; Unbedingtes Verlassen einer DO-LOOP-Schleife
; mit Beseitigen der Parameter vom Return-Stack.
;------------------------------------------------------------------------------

NFA_LEAVE:	db	85h
		db	"LEAVE"
		dw	NFA_IDO
		db	0
;
LEAVE:		pop	R8
		pop	R9
		pop	R4
		pop	R5
		pop	R6
		pop	R7
		push	R5
		push	R4
		push	R5
		push	R4
		jp	@RR8

;------------------------------------------------------------------------------
; BEGIN ( -- )
; Erffnung einer Programmschleife der Struktur BEGIN - UNTIL
; oder BEGIN - WHILE - REPEAT
;------------------------------------------------------------------------------

NFA_BEGIN:	db	85h
		db	"BEGIN"
		dw	NFA_LEAVE
		db	80h
;
BEGIN:		ret

;------------------------------------------------------------------------------
; UNTIL (? -- )
; Ist Flag ? wahr, erfolgt die Fortsetzung mit dem folgenden Wort,
; sonst Rcksprung zu BEGIN. Schleifenabschlu.
;------------------------------------------------------------------------------

NFA_UNTIL:	db	85h
		db	"UNTIL"
		dw	NFA_BEGIN
		db	22h
;
UNTIL:		call	DROP
		pop	R8
		pop	R9
		ld	R10, #16h	; 16h = R6
		ldci	@R10, @RR8
		ldci	@R10, @RR8
		or	R4, R5
		jr	Z, until1
		jp	@RR8
until1:		jp	@RR6

;------------------------------------------------------------------------------
; WHILE ( ? -- )
; Ist Flag ? wahr. erfolgt die Fortsetzung mit dem folgenden Wort,
; sonst mit dem ersten Wort hinter REPEAT.
;------------------------------------------------------------------------------

NFA_WHILE:	db	85h
		db	"WHILE"
		dw	NFA_UNTIL
		db	0C2h
;
WHILE:		jr	UNTIL

;------------------------------------------------------------------------------
; REPEAT ( -- )
; Programmfortsetzung mit dem ersten Wort nach BEGIN Schleifenabschlu.
;------------------------------------------------------------------------------

NFA_REPEA:	db	86h
		db	"REPEAT"
		dw	NFA_WHILE
		db	32h
;
REPEA:		pop	R8
		pop	R9
		ld	R10, #16h	; 16h = R6
		ldci	@R10, @RR8
		ldci	@R10, @RR8
		jp	@RR6

;------------------------------------------------------------------------------
; AGAIN
;------------------------------------------------------------------------------

NFA_AGAIN:	db	85h
		db	"AGAIN"
		dw	NFA_REPEA
		db	22h
AGAIN:		jr	REPEA

;------------------------------------------------------------------------------
; KEY (-- 16b)
; Die niederen sieben Bit von 16b sind das ASCII der bettigten Taste.
;------------------------------------------------------------------------------

NFA_KEY:	db	83h
		db	"KEY"
		dw	NFA_AGAIN
		db	0
;
KEY:		call	TSK		; ES 4.0 Tastenabfrage
		ld	R4, #0
		ld	R5, 5Ah
		jp	TOSTACK

;------------------------------------------------------------------------------
; CR (--)
; Ausgabe des Steuerzeichens Zeilenschaltung (carriage return und line feed).
;------------------------------------------------------------------------------

NFA_CR:		db	82h
		db	"CR"
		dw	NFA_KEY
		db	0
;
CR:		jp	NLE	; ES: Zeilenschaltung

;------------------------------------------------------------------------------
; EMIT (16b --)
; Ausgabe von 16b als ASCII auf dem Bildschirm.
; Meist werden nur die sieben niederwertigen Dualstellen verwendet.
;------------------------------------------------------------------------------

NFA_EMIT:	db	84h
		db	"EMIT"
		dw	NFA_CR
		db	0
;
EMIT:		call	DROP
		ld	5Ah, R5
		jp	ZBS        ; Ausgabe %5A

;------------------------------------------------------------------------------
; KEY? ( -- ?)
; dynamische Tastenabfrage, kein Warten
; Flag ist bei Nichtbettigung gleich Null (falsch)
;------------------------------------------------------------------------------

NFA_KEYQ:	db	84h
		db	"KEY?"
		dw	NFA_EMIT
		db	0
;
KEYQ:		call	CKEY		; ES: Tastenabfrage
		ld	R4, #0
		ld	R5, 6Dh
		jp	TOSTACK

;patch CR, Zeilenschaltung
NLE:       	ld      5Ah, #0Dh
		nop
		jp      ZBS        ; Ausgabe %5A

;------------------------------------------------------------------------------
; DP
; Systemvariable (dictionary pointer) 
;------------------------------------------------------------------------------

NFA_DP:		db	82h
		db	"DP"
		dw	NFA_KEYQ
		db	0
;
DP:		call	VAR
		dw	INITDP

;------------------------------------------------------------------------------
; HERE (-- adr)
; adr ist die Adresse der nchsten verfgbaren Speicherzelle im Wrterbuch.
;------------------------------------------------------------------------------

		db	0FFh		; ???

NFA_HERE:	db	84h
		db	"HERE"
		dw	NFA_DP
		db	0
;
HERE:		call	DP
		jr	AT

;------------------------------------------------------------------------------
; @ (adr -- 16b)
; 16b ist der Inhalt der Speicherzellen adr und adr+1.
;------------------------------------------------------------------------------

NFA_AT:		db	81h
		db	'@'
		dw	NFA_HERE
		db	0
;
AT:		ldc	R4, @RR2
		incw	RR2
		ldc	R3, @RR2
		ld	R2, R4
		ret

;------------------------------------------------------------------------------
; 2+
;------------------------------------------------------------------------------

NFA_TWOP:	db	82h
		db	"2+"
		dw	NFA_AT
		db	0
;
TWOP:		incw	RR2
		incw	RR2
		ret

;------------------------------------------------------------------------------
; ENCLOSE
; The text scanning primitive
;------------------------------------------------------------------------------

NFA_ENCL:	db	87h
		db	"ENCLOSE"
		dw	NFA_TWOP
		db	0
;
ENCL:		ld	R0, #0
		ld	R1, #0
		ld	R4, R2
		ld	R5, R3
		ld	R7, #20h
encl1:		lde	R6, @RR4
		or	R6, R6
		jr	Z, encl3
		ld	R1, R5
		call	encl4
		cp	R6, R7
		jr	Z, encl1
encl2:		lde	R6, @RR4
		or	R6, R6
		jr	Z, encl3
		inc	R0
		call	encl4
		cp	R6, R7
		jr	NZ, encl2
encl3:		dec	R1
		ld	R3, R5
		ld	R5, R1
		lde	@RR4, R0
		ret
;
encl4:		inc	R5
		nop			;patched
		nop
		nop
		nop
		nop
		nop
		nop
		nop
		ret

;------------------------------------------------------------------------------
; -FIND
; searches for a matching entry
;------------------------------------------------------------------------------

NFA_MFIND:	db	85h
		db	"-FIND"
		dw	NFA_ENCL
		db	0
;
MFIND:		call	HERE
mfind1:		ldc	R8, @RR2
		incw	RR2
		ldc	R9, @RR2
		ld	R3, R8
		or	R3, R9
		ld	R2, #0
		jr	NZ, mfind2
		ret
;
mfind2:		ld	R4, #hi(TIBBUF)	; ! F3
		ld	R5, R1
		lde	R6, @RR4
		ldc	R7, @RR8
		and	R7, #7Fh
		cp	R6, R7
		jr	Z, mfind4
mfind3:		scf
		adc	R9, R7
		adc	R2, R8
		ld	R3, R9
		jr	mfind1
;
mfind4:		call	encl4
mfind5:		incw	RR8
		lde	R10, @RR4
		ldc	R11, @RR8
		call	encl4
		dec	R7
		cp	R10, R11
		jr	NZ, mfind3
		or	R7, R7
		jr	NZ, mfind5
		add	R9, #3
		adc	R8, R7
		ldc	R14, @RR8
		ld	R2, R8
		ld	R3, R9
		incw	RR2
		ret

;------------------------------------------------------------------------------
; EXPECT (adr +n --)
; Eingabe von hchstens +n Zeichen und Abspeichern ab adr.
; Beim Erkennen des ENTER-Zeichens Rckkehr ohne Speichern des ENTER-Codes.
; Speichern eines zustzlichen Leerzeichens.
; Anzeige aller Zeichen auer ENTER auf dem Bildschirm (Echo).
;------------------------------------------------------------------------------

NFA_EXPECT:	db	86h
		db	"EXPECT"
		dw	NFA_MFIND
		db	0

;
EXPEC:		jp      expec0
;patched (expect ausgelagert)
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh

;------------------------------------------------------------------------------
; INTERPRET
;------------------------------------------------------------------------------

NFA_INTER:	db	89h
		db	"INTERPRET"
		dw	NFA_EXPECT
		db	0
INTER:		call	SCSP
inter1:		call	TIB
		call	AT
		call	sub_E14A
		call	DROP
		or	R4, R5
		jr	NZ, inter2
		call	sub_E13D
inter2:		call	ENCL
		or	R0, R0
		jr	NZ, inter5
		call	QCSP
inter3:		call	TIB
		call	STORE
		call	STATE
		call	AT
		call	DROP
		or	R4, R5
		jr	Z, inter4
		ret

		db	0FFh		; ???

inter4:		ld	R6, #0AFh	; RET
		ldc	@RR12, R6
		call	HERE
		call	TWOP
		call	DROP
		jp	@RR4
;
inter5:		call	MFIND
		call	DROP
		ld	R6, R4
		or	R6, R5
		jr	NZ, inter9
		call	NUMB
		call	dpl1
		call	AT
		call	DROP
		call	inter21
		ld	R6, #hi(DLIT)	; ! und hi(LIT)
		ldc	@RR12, R6
		incw	RR12
		incw	RR4
		jr	Z, inter6
		ld	R6, #lo(DLIT)	; !
		ldc	@RR12, R6
		incw	RR12
		ld	R6, #12h	; ? s.a. LIT/DLIT
		ldci	@RR12, @R6
		ldci	@RR12, @R6
		jr	inter7
;
inter6:		ld	R6, #lo(LIT)	; !
		ldc	@RR12, R6
		incw	RR12
inter7:		call	DROP
		ld	R6, #12h	; 12h = RR2 = TOS
		ldci	@RR12, @R6
		ldci	@RR12, @R6
		call	DROP
inter8:		call	TIB
		call	AT
		jr	inter2
;
inter9:		tm	R14, #1
		jr	Z, inter10
		push	R5
		push	R4
		call	inter3
		pop	R4
		pop	R5
		call	@RR4
		jp	inter1
;
inter10:	call	inter21
		ld	R6, #14h	; 14h = RR4
		ldci	@RR12, @R6
		ldci	@RR12, @R6
		ld	R10, R12
		ld	R11, R13
		ld	R6, R14
		and	R6, #6
		jr	Z, inter16
		cp	R6, #2
		jr	Z, inter11
		cp	R6, #4
		jr	NZ, inter12
		incw	RR12
		incw	RR12
inter11:	incw	RR12
		incw	RR12
		jr	inter16
;
inter12:	clr	R7
inter13:	incw	RR12
		lde	R6, @RR2
		inc	R3
		nop        		;patched
		nop
		nop
		nop
		nop
		nop
		nop
		nop
inter14:	cp	R6, #22h
		jr	Z, inter15
		ldc	@RR12, R6
		inc	R7
		jr	inter13
;
inter15:	ldc	@RR10, R7
		nop			; ???
inter16:	tm	R14, #10h
		jr	Z, inter17
		pop	R8
		pop	R9
		ldc	@RR8, R12
		incw	RR8
		ldc	@RR8, R13
inter17:	tm	R14, #20h
		jr	Z, inter18
		pop	R8
		pop	R9
		ldc	@RR10, R8
		incw	RR10
		ldc	@RR10, R9
		decw	RR10
inter18:	tm	R14, #40h
		jr	Z, inter19
		push	R11
		push	R10
inter19:	tm	R14, #80h
		jr	Z, inter20
		push	R13
		push	R12
inter20:	jp	inter2
;
inter21:	ld	R6, #0D6h	; CALL
		ldc	@RR12, R6
		incw	RR12
		ret

;------------------------------------------------------------------------------
; NUMBER
; f. INPUT
;------------------------------------------------------------------------------

NFA_NUMB:	db	86h
		db	"NUMBER"
		dw	NFA_INTER
		db	0
;
NUMB:		call	TIB
		call	STORE
		call	TOSTACK
		ld	R0, #0
		ld	R2, #0
		ld	R3, #0
		ld	R4, #0
		ld	R5, #0
		ld	R14, #0
		call	TOSTACK
		call	TOSTACK
		ld	R2, #hi(TIBBUF)	; ! F3
		ld	R3, R1
		ld	R4, #0FFh
		ld	R5, #0FFh
		call	TOSTACK
		call	dpl1
		call	STORE
numb1:		inc	R3
		nop			; patched
		nop
		nop
		nop
		nop
		nop
		nop
		nop
numb2:		ldc	R7, @RR2
		cp	R7, #' '
		jr	Z, numb8
		cp	R7, #'.'
		jr	Z, numb7
		cp	R7, #'-'
		jr	NZ, numb3
		ld	R14, #0FFh
		jr	numb1
;
numb3:		or	R7, R7
		jr	Z, numb8
		or	R0, R0
		jr	Z, numb4
		call	dpl1
		call	AT
		incw	RR2
		call	dpl1
		call	STORE
numb4:		sub	R7, #'0'
		jr	C, numb9
		cp	R7, #0Ah
		jr	C, numb5
		sub	R7, #7
		cp	R7, #0Ah
		jr	C, numb9
numb5:		call	ROT
		call	ROT
		call	BASE
		call	AT
		cp	R7, R3
		jr	NC, numb9
		push	R7
		call	DSTAR
		call	DROP
		ld	R6, #0
		pop	R7
		add	R3, R7
		adc	R2, R6
		adc	R5, R6
		adc	R4, R6
		call	TOSTACK
		call	ROT
numb6:		jr	numb1
;
numb7:		inc	R0
		ld	R4, #0
		ld	R5, #0
		call	TOSTACK
		call	DPL
		call	STORE
		jr	numb6
;
numb8:		call	DROP
		inc	R14
		jr	Z, DNEG
		ret
;
numb9:		ld	R3, R1
		ld	R2, #hi(TIBBUF)	; ! F3
		incw	RR2
		nop
numb10:		ld	R6, #5Ah
		ldci	@R6, @RR2
		cp	5Ah, #20h
		jr	Z, numb11
		call    ZBS        ; Ausgabe %5A
		nop
		nop
		nop
		nop
		nop
		nop
		nop
		nop
		jr	numb10
;
numb11:		ld	5Ah, #3Fh
		call    ZBS        ; Ausgabe %5A
		ld	R6, #'0'
		jp	error		; Ausgabe Fehler #MSG <R6>

;------------------------------------------------------------------------------
; DNEGATE (d1 -- d2)
; d2 ist die Differenz 0 minus d1.
;------------------------------------------------------------------------------

NFA_DNEG:	db	87h
		db	"DNEGATE"
		dw	NFA_NUMB
		db	0
;
DNEG:		call	DROP
		com	R2
		com	R3
		com	R4
		com	R5
		incw	RR2
		jr	NZ, dneg1
		incw	RR4
dneg1:		jp	TOSTACK

;------------------------------------------------------------------------------
; DABS (d -- ud)
; ud ist der Absolutbetrag von d.
;------------------------------------------------------------------------------

NFA_DABS:	db	84h
		db	"DABS"
		dw	NFA_DNEG
		db	0
;
DABS:		nop			; ???
		or	R2, R2
		jr	MI, DNEG
		ret

;------------------------------------------------------------------------------
; D*
;------------------------------------------------------------------------------

NFA_DSTAR:	db	82h
		db	"D*"
		dw	NFA_DABS
		db	0

DSTAR:		call	SWAP
		call	OVER
		call	USTAR
		call	DROP
		call	ROT
		call	ROT
		call	USTAR
		call	ROT
		jp	PLUS

;------------------------------------------------------------------------------
; ?CSP
;------------------------------------------------------------------------------

NFA_QCSP:	db	84h
		db	"?CSP"
		dw	NFA_DSTAR
		db	0
;
QCSP:		call	rpat1
		call	CSP
		call	AT
		call	DROP
		ld	R6, #'J'
		jr	qpair1

;------------------------------------------------------------------------------
; ?PAIRS
;------------------------------------------------------------------------------

NFA_QPAIR:	db	86h
		db	"?PAIRS"
		dw	NFA_QCSP
		db	0
;
		call	DROP
		ld	R6, #'I'
qpair1:		sub	R3, R5
		sbc	R2, R4
		or	R2, R3
		jp	NZ, error	; Ausgabe Fehler #MSG <R6>
		jp	DROP

;------------------------------------------------------------------------------
; DPL
;------------------------------------------------------------------------------

nFA_DPL:	db	83h
		db	"DPL"
		dw	NFA_QPAIR
		db	0
;
DPL:		nop			; ???
dpl1:		call	VAR
		dw	0FFFFh

;------------------------------------------------------------------------------
; !CSP
;------------------------------------------------------------------------------

NFA_SCSP:	db	84h
		db	"!CSP"
		dw	nFA_DPL
		db	0
;
SCSP:		call	rpat1
		call	CSP
		jp	STORE

;------------------------------------------------------------------------------
; BASE (-- adr)
; Adresse der Systemvariablen zum Speichern der Zahlenbasis fr die Ein- und Ausgabe.
;------------------------------------------------------------------------------

NFA_BASE:	db	84h
		db	"BASE"
		dw	NFA_SCSP
		db	0
;
BASE:		call	VAR
		dw	10

;------------------------------------------------------------------------------
; RP@
;------------------------------------------------------------------------------

NFA_RPAT:	db	83h
		db	"RP@"
		dw	NFA_BASE
		db	0
;
RPAT:		nop			; ???
rpat1:		ld	R5, spl		; Stack	pointer
		ld	R4, sph		; Stack pointer
		incw	RR4
		incw	RR4
		jp	TOSTACK

;------------------------------------------------------------------------------
; CSP
;------------------------------------------------------------------------------

NFA_CSP:	db	83h
		db	"CSP"
		dw	NFA_RPAT
		db	0
;
CSP:		call	VAR
		dw	RPSTACK-4

;------------------------------------------------------------------------------
; WORDS
; Wortschatz auflisten. Pause/Weiter mit ENTER
;------------------------------------------------------------------------------

NFA_WORDS:	db	85h
		db	"WORDS"
		dw	NFA_CSP
		db	0
;
WORDS:		call	DP
		call	AT
words1:		ld	R5, #7
words2:		ldc	R8, @RR2
		incw	RR2
		ldc	R9, @RR2
		ld	R4, R8
		or	R4, R9
		jp	Z, DROP
		ldc	R7, @RR8
		incw	RR8
		and	R7, #7Fh
words3:		ld	R6, #5Ah
		ldci	@R6, @RR8
		call    ZBS        ; Ausgabe %5A
		djnz	R7, words3
		call    SPACE
		call    SPACE
		ld	R2, R8
		ld	R3, R9
		djnz	R5, words2
		call	TSK		; ES 4.0: Tastaturabfrage
		cp	5Ah, #0Dh
		jr	Z, words1
		jp	DROP

;------------------------------------------------------------------------------
; STATE (-- adr)
; Adresse einer Variablen, die den ausfhrenden Modus mit 0 und den
; compilierenden mit einem Inhalt ungleich 0 kennzeichnet.
;------------------------------------------------------------------------------

NFA_STATE:	db	85h
		db	"STATE"
		dw	NFA_WORDS
		db	0
;
STATE:		call	VAR
		dw	0

;------------------------------------------------------------------------------
; CREATE
;------------------------------------------------------------------------------

NFA_CREAT:	db	86h
		db	"CREATE"
		dw	NFA_STATE
		db	0
;
CREAT:		call	HERE
		call	DUP
		call	STATE
		call	STORE
		ldc	R10, @RR2
		incw	RR2
		ldc	R11, @RR2
		decw	RR2
		ld	R12, R2
		ld	R13, R3
		call	DROP
		call	TIB
		call	AT
		call	ENCL
		ld	R6, #'D'
		or	R0, R0
		jr	Z, creat2
		ld	R6, R0
		or	R6, #80h
		ldc	@RR12, R6
		incw	RR12
creat1:		call	encl4
		lde	R6, @RR4
		ldc	@RR12, R6
		incw	RR12
		djnz	R0, creat1
		ld	R4, R12
		ld	R5, R13
		call	TOSTACK
		call	DP
		call	STORE
		ld	R6, #1Ah
		ldci	@RR12, @R6
		ldci	@RR12, @R6
		clr	R6
		ldc	@RR12, R6
		incw	RR12
		call	TIB
		call	STORE
		ret
;
creat2:		push	R6
		call	HERE
		call	AT
		call	STATE
		call	AT
		call	DUP
		call	DP
		call	STORE
		call	STORE
		ld	R4, #0
		ld	R5, #0
		call	TOSTACK
		call	STATE
		call	STORE
		pop	R6
		jp	error		; Ausgabe Fehler #MSG <R6>

;------------------------------------------------------------------------------
; VARIABLE name
; Variablendefinition
;------------------------------------------------------------------------------

NFA_VARIAB:	db	88h
		db	"VARIABLE"
		dw	NFA_CREAT
		db	1
;
		call	CREAT
		ld	R8, #hi(VAR)
		ld	R9, #lo(VAR)
		call	variab2
		incw	RR12
		incw	RR12
variab1:	ld	R4, R12
		ld	R5, R13
		incw	RR12
		incw	RR12
		call	TOSTACK
		call	STATE
		call	AT
		call	OVER
		call	STORE
		call	DP
		call	STORE
		ld	R4, #0
		ld	R5, #0
		call	TOSTACK
		call	STATE
		jp	STORE
;
variab2:	call	inter21
		ld	R6, #18h
		ldci	@RR12, @R6
		ldci	@RR12, @R6
		ret

;------------------------------------------------------------------------------
; CONSTANT name (n --)
; Konstantendefinition
;------------------------------------------------------------------------------

NFA_CONST:	db	88h
		db	"CONSTANT"
		dw	NFA_VARIAB
		db	1
;
		call	CREAT
		ld	R8, #hi(CON)
		ld	R9, #lo(CON)
		call	variab2
		ld	R6, #12h	; 12h = RR2 = TOS
		ldci	@RR12, @R6
const1:		ldci	@RR12, @R6
		call	DROP
		jr	variab1

;------------------------------------------------------------------------------
; ;
;------------------------------------------------------------------------------

NFA_SEMIS:	db	81h
		db	';'
		dw	NFA_CONST
		db	1
;
SEMIS:		call	rpat1
		call	CSP
		call	AT
		call	DROP
		sub	R3, R5
		sbc	R2, R4
		or	R2, R3
		ld	R6, #'I'
		jp	NZ, creat2
		call	sub_E16B
		ld	R6, #0AFh
		ldc	@RR12, R6
		incw	RR12
		jp	variab1

;------------------------------------------------------------------------------
; :
;------------------------------------------------------------------------------

NFA_COLON:	db	81h
		db	':'
		dw	NFA_SEMIS
		db	1
;
COLON:		call	CREAT
		pop	R4
		pop	R5
		call	SCSP
		jp	inter8

;------------------------------------------------------------------------------
; FORGET name
; entfernt name und alle folgenden  Eintrge aus dem Wrterbuch
;------------------------------------------------------------------------------

		db	0FFh		; ???
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh
		db	0FFh

NFA_FORGT:	db	86h
		db	"FORGET"
		dw	NFA_COLON
		db	1
;
		call	TIB
		call	AT
		call	ENCL
		ld	R6, #'D'
		or	R0, R0
		jp	Z, error	; Ausgabe Fehler #MSG <R6>
		call	MFIND
		ld	R6, R2
		or	R6, R3
		jp	Z, numb9
		call	NFA
		call	DUP
		call	LFA
		call	AT
		call	OVER
		call	STORE
		call	DP
		call	STORE
		call	TIB
		jp	STORE

;------------------------------------------------------------------------------
; NFA
;------------------------------------------------------------------------------

NFA_NFA:	db	83h
		db	"NFA"
		dw	NFA_FORGT
		db	0
;
NFA:		sub	R3, #4
		ld	R4, #0
		sbc	R2, R4
nfa1:		ldc	R5, @RR2
		or	R5, R5
		jr	MI, nfa2
		inc	R4
		decw	RR2
		jr	nfa1
;
nfa2:		and	R5, #7Fh
		cp	R4, R5
		jr	Z, nfa3
		clr	R2
		clr	R3
nfa3:		ret

;------------------------------------------------------------------------------
; LFA
;------------------------------------------------------------------------------

NFA_LFA:	db	83h
		db	"LFA"
		dw	NFA_NFA
		db	0
;
LFA:		ldc	R5, @RR2
		and	R5, #7Fh
		scf
		adc	R3, R5
		adc	R2, #0
		ret

;------------------------------------------------------------------------------
; .!
;------------------------------------------------------------------------------

; patched
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh
		db 	0FFh


;------------------------------------------------------------------------------
; CODE
;------------------------------------------------------------------------------

NFA_CODE:	db	84h
		db	"CODE"
		dw	NFA_LFA
		db	1
;
		call	CREAT
		ld	20h, R12
		ld	21h, R13
		call	ES_DATA		; ES: DATA-Modus
		srp	#10h
		ld	R12, 20h
		ld	R13, 21h
		call	TIB
		call	AT
		ld	R4, #0
		lde	@RR2, R4
		call	DROP
		jp	variab1

;------------------------------------------------------------------------------
; CFA
;------------------------------------------------------------------------------

NFA_CFA:	db	83h
		db	"CFA"
		dw	NFA_CODE
		db	0
;
		ldc	R5, @RR2
		and	R5, #7Fh
		add	R5, #3
		add	R3, R5
		adc	R2, #0
		ret

;------------------------------------------------------------------------------
; ID.
;------------------------------------------------------------------------------

NFA_IDDOT:	db	83h
		db	"ID."
		dw	NFA_CFA
		db	0
;
IDDOT:		ldc	R5, @RR2
		and	R5, #7Fh
		incw	RR2
iddot1:		ld	R4, #5Ah
		ldci	@R4, @RR2
		call    ZBS        ; Ausgabe %5A
		djnz	R5, iddot1
		call	SPACE
		jp	DROP

;------------------------------------------------------------------------------
; SPACE (--)
; Ausgabe eines Leerzeichens.
;------------------------------------------------------------------------------

NFA_SPACE:	db	85h
		db	"SPACE"
		dw	NFA_IDDOT
		db	0
;
SPACE:		ld	5Ah, #' '
		jp	ZBS	; Ausgabe %5A

;------------------------------------------------------------------------------
; SPACES
;------------------------------------------------------------------------------

NFA_SPACES:	db	86h
		db	"SPACES"
		dw	NFA_SPACE
		db	0
SPACES:		call	DROP
spaces1:	call	SPACE
		djnz	R5, spaces1
		ret

;------------------------------------------------------------------------------
; #> (32b -- adr +n)
; Abschlu der Zeichenkette im Ausgabe-Stapelspeicher,
; entfernen von 32b vom Stack und bergabe von Adresse adr und Lnge +n
; fr die Anzeige mit TYPE.
;------------------------------------------------------------------------------


		dw	0F9AFh		; ???

NFA_EDIGS:	db	82h
		db	"#>"
		dw	NFA_SPACES
		db	0
;
EDIGS:		call	DROP
		call	DROP
		call	HLD
		call	AT
		; statt PAD OVER MINUS
		ld	R5, R3
		ld	R4, #0
		com	R5              ; Lnge, deswegen xx00-Adresse
		inc	R5
		jp	TOSTACK

;------------------------------------------------------------------------------
; TYPE (adr +n --)
; Ab adr werden +n Zeichen aus dem Speicher gelesen
; und auf dem Bildschirm ausgegeben.
;------------------------------------------------------------------------------

NFA_TYPE:	db	84h
		db	"TYPE"
		dw	NFA_EDIGS
		db	0
;
TYPE:		call	DROP
type1:		ld	R6, #5Ah
		ldci	@R6, @RR2
		call    ZBS        ; Ausgabe %5A
		djnz	R5, type1
		jp	DROP

;------------------------------------------------------------------------------
; HOLD (16b --)
; Eintragen von 16b als ASCII in den Ausgabe-Stapelspeicher.
;------------------------------------------------------------------------------

NFA_HOLD:	db	84h
		db	"HOLD"
		dw	NFA_TYPE
		db	0
;
HOLD:		call	HLD
		call	AT
		call	DROP
		decw	RR4
		ldc	@RR4, R3
		call	sub_E110
		call	HLD
		jp	STORE

;------------------------------------------------------------------------------
;
; SIGN (n --)
; Eintragen des ASCII fr - (minus) in den Ausgabe-Stapelspeicher,
; wenn n negativ ist.
; ACHTUNG JU+TE-FORTH: funktioniert wie in FIG, nicht wie in FORTH 83
;------------------------------------------------------------------------------

NFA_SIGN:	db	84h
		db	"SIGN"
		dw	NFA_HOLD
		db	0
;
SIGN:		call	ROT
		or	R2, R2
		jp	PL, DROP
		ld	R3, #2Dh
		jr	HOLD

;------------------------------------------------------------------------------
; S->D ( n -- d)
;------------------------------------------------------------------------------

NFA_STOD:	db	84h
		db	"S->D"
		dw	NFA_SIGN
		db	0
;
STOD:		ld	R4, #0FFh
		ld	R5, #0FFh
		or	R2, R2
		jr	MI, stod1
		incw	RR4
stod1:		jp	TOSTACK

;------------------------------------------------------------------------------
; <# (--)
; Initialisieren der Zahlenausgabe mit Erffnen eines Stapelspeichers
; fr die auszugebenden Zeichen.
;------------------------------------------------------------------------------

NFA_BDIGS:	db	82h
		db	"<#"
		dw	NFA_STOD
		db	0
;
BDIGS:		ld	R4, #hi(MDIGS)
		ld	R5, #lo(MDIGS)
		call	TOSTACK
		call	HLD
		jp	STORE

;------------------------------------------------------------------------------
; HLD
; Adresse fr Zeichenstapel fr Zahlkonvertierung
;------------------------------------------------------------------------------

NFA_HLD:	db	83h
		db	"HLD"
		dw	NFA_BDIGS
		db	0
;
HLD:		call	VAR
		dw	MDIGS-2

;------------------------------------------------------------------------------
; # (+d1 -- +d2)
; Division von +d1 durch die in BASE gespeicherte Zahlenbasis.
; +d2 ist der Quotient, der Divisionsrest wird als ASCII
; in den Ausgabe-Stapelspeicher eingetragen.
;------------------------------------------------------------------------------

NFA_DIG:	db	81h
		db	'#'
		dw	NFA_HLD
		db	0
;
DIG:		call	BASE
		call	AT
		call	MSMOD
		call	ROT
		call	DROP
		ld	R7, R5
		cp	R7, #0Ah
		jr	C, dig1
		add	R7, #7
dig1:		add	R7, #'0'
		ld	R5, R7
		ld	R4, #0
		call	TOSTACK
		jp	HOLD

;------------------------------------------------------------------------------
; #S (+d -- 0 0)
; Zyklisches Ausfhren von #, bis der doppelt genaue Quotient gleich Null ist.
;------------------------------------------------------------------------------

NFA_DIGS:	db	82h
		db	"#S"
		dw	NFA_DIG
		db	0
;
DIGS:		call	DIG
		call	DROP
		ld	R7, R2
		or	R7, R3
		call	TOSTACK
		or	R7, R2
		or	R7, R3
		jr	NZ, DIGS
		ret

;------------------------------------------------------------------------------
; U. (u --)
; Zahlenausgabe in vorzeichenloser Deutung. Doppelt genaue Zahlenausgaben
;------------------------------------------------------------------------------

NFA_UDOT:	db	82h
		db	"U."
		dw	NFA_DIGS
		db	0
;
		call	LIT
		dw	0
		call	BDIGS
		call	DIGS
		call	EDIGS
		jp	sub_E189
		ret

;------------------------------------------------------------------------------
; . (n --)
; Ausgabe des TOS als Zahl bezglich der in BASE gespeicherten Basis.
; Negative Werte kennzeichnet ein vorangestelltes Minuszeichen.
;------------------------------------------------------------------------------

NFA_DOT:	db	81h
		db	'.'
		dw	NFA_UDOT
		db	0
;
DOT:		call	DUP
		call	STOD
dot1:		call	DABS
		call	BDIGS
		call	DIGS
		call	SIGN
		call	EDIGS
		call	sub_E189
		ret

;------------------------------------------------------------------------------
; D. (d --)
; Ausgabe von d als vorzeichenbehaftete Zahl.
;------------------------------------------------------------------------------

NFA_DDOT:	db	82h
		db	"D."
		dw	NFA_DOT
		db	0
;
DDOT:		call	SWAP
		call	OVER
		jr	dot1

		ret			; ???

;------------------------------------------------------------------------------
; anstelle VIEW 
;------------------------------------------------------------------------------

		db	85h		; ???

; zu query (patch)
query0:         call    TOSTACK
                ld      055h, #57h
                ld      R6, #3
                ld      R4, #hi(TIBBUF)	; ! F3
                ld      R5, #0
                ld      R2, #20h
query1:         lde     @RR4, R2	; Buffer mit Leerzeichen fllen
                djnz    R5, query1
                inc     R5
query2:         ld      058h, #0FFh	; neue Zeile init
                call    CHARIN	; ES:
                ld      R2, #0F7h	; CHARIN legt Eingabezeile
                ld      R3, #00h	; auf F700 ff. ab
                ld      R1, #4Fh	; ?
query3:         lde     R0, @RR2
                lde     @RR4, R0
                inc     R5
                inc     R3
                djnz    R1, query3
                lde     R0, @RR2
                cp      R0, #5Ch
                jr      NZ, query4
                djnz    R6, query2
                ld      R6, #5Ch
                jp      error
;
query4:         ld      R5, #0FFh
                ld      R3, #0
                lde     @RR4, R3
                ld      R2, #hi(TIBBUF)	; ! F3
                call    TIB
                call    STORE
                ret
; ---------------------------------------------------------------------------
                db	0FFh
                db	0FFh
                db	0FFh
; ---------------------------------------------------------------------------
; expect (patch)
expec0:         ld      058h, #0FFh	; neue Input-Zeile init.
                call    DROP
expec1:         push    R3
                call    CHARIN	; ES:
                ld      R6, R3
                pop     R3
                lde     @RR2, R6
                incw    RR2
                cp      R6, #0Dh	; Eingabe-Ende
                jr      Z, expec2
                decw    RR4
                jr      NZ, expec1
;
expec2:         decw    RR2
                ld      R6, #0
                lde     @RR2, R6
                jp      DROP
; ---------------------------------------------------------------------------
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh
                db	0FFh

;------------------------------------------------------------------------------
; SP@
;------------------------------------------------------------------------------

NFA_SPAT:	db	83h
		db	"SP@"
		dw	NFA_DDOT
		db	0
;
SPAT:		ld	R5, R15		; SP
		ld	R4, #0
		jp	TOSTACK

;------------------------------------------------------------------------------
; SP!
;------------------------------------------------------------------------------

NFA_SPSTOR:	db	83h
		db	"SP!"
		dw	NFA_SPAT
		db	0
;
		ld	R15, #50h	; SP
		ret

;------------------------------------------------------------------------------
; RP!
;------------------------------------------------------------------------------

NFA_RPSTOR:	db	83h
		db	"RP!"
		dw	NFA_SPSTOR
		db	0
;
		pop	R4
		pop	R5
		clr	spl		; Stack	pointer
		ld	sph, #hi(RPSTACK)	; Stack pointer
		jp	@RR4

;------------------------------------------------------------------------------
; AND (16b1 16b2 -- 16b3)
; 16b3 ist die bitweise logische AND-Verknpfung von 16b1 und 16b2 (UND).
;------------------------------------------------------------------------------

NFA_ANDD:	db	83h
		db	"AND"
		dw	NFA_RPSTOR
		db	0
;
		call	DROP
		and	R2, R4
		and	R3, R5
		ret

;------------------------------------------------------------------------------
; OR (16b1 16b2 -- 16b3)
; 16b3 ist die bitweise logische OR-Verknpfung von 16b1 und 16b2 (ODER).
;------------------------------------------------------------------------------

NFA_ORR:	db	82h
		db	"OR"
		dw	NFA_ANDD
		db	0
;
		call	DROP
		or	R2, R4
		or	R3, R5
		ret

;------------------------------------------------------------------------------
; XOR (16b1 16b2 -- 16b3)
; 16b3 ist die bitweise logische XOR-Verknpfung von 16b1 und 16b2 (ANTIVALENZ).
;------------------------------------------------------------------------------

NFA_XORR:	db	83h
		db	"XOR"
		dw	NFA_ORR
		db	0
;
		call	DROP
		xor	R2, R4
		xor	R3, R5
		ret

;------------------------------------------------------------------------------
; = (wt w2 -- ?)
; Flag ? ist wahr, wenn w1 = w2.
;------------------------------------------------------------------------------


NFA_EQUAL:	db	81h
		db	'='
		dw	NFA_XORR
		db	0
;
		call	DROP
		sub	R3, R5
		sbc	R2, R4
equal1:		or	R2, R3
		clr	R2
		clr	R3
		jr	NZ, equal2
		decw	RR2
equal2:		ret

;------------------------------------------------------------------------------
; > (n1 n2 -- ?)
; Flag ? ist wahr, wenn n1 > n2.
;------------------------------------------------------------------------------

NFA_GREAT:	db	81h
		db	'>'
		dw	NFA_EQUAL
		db	0
;
		call	DROP
great1:		sub	R5, R3
		sbc	R4, R2
		clr	R2
		clr	R3
		jr	GE, equal2
		decw	RR2
		ret

;------------------------------------------------------------------------------
; < (n1 n2 -- ?)
; Flag ? ist wahr, wenn n1 < n2.
;------------------------------------------------------------------------------

NFA_LESS:	db	81h
		db	'<'
		dw	NFA_GREAT
		db	0
;
		call	DROP
less1:		sub	R3, R5
		sbc	R2, R4
		clr	R2
		clr	R3
		jr	GE, equal2
		decw	RR2
		ret

;------------------------------------------------------------------------------
; 0= (w -- ?)
; Flag ? ist wahr, wenn w = 0.
; Dieses Wort eignet sich zum Negieren eines Flags.
;------------------------------------------------------------------------------


NFA_ZEQUAL:	db	82h
		db	"0="
		dw	NFA_LESS
		db	0
;
		jr	equal1

;------------------------------------------------------------------------------
; 0< (n--?)
; Flag ? ist wahr, wenn n < 0.
;------------------------------------------------------------------------------

NFA_ZLESS:	db	82h
		db	"0<"
		dw	NFA_ZEQUAL
		db	0
;
		clr	R4
		clr	R5
		jr	less1

;------------------------------------------------------------------------------
; 0> (n -- ?)
; Flag ? ist wahr, wenn n > 0.
;------------------------------------------------------------------------------

NFA_ZGREAT:	db	82h
		db	"0>"
		dw	NFA_ZLESS
		db	0
;
ZGREAT:		clr	R4
		clr	R5
		jr	great1

;------------------------------------------------------------------------------
; IF (? -- )
; Beginn einer Verzweigung. Ist ? wahr belegt, werden die folgenden Wrter
; bis ELSE bzw. THEN ausgefhrt. Ist ? falsch belegt, kommen die Wrter
; zwischen ELSE und THEN zur Ausfhrung.
;------------------------------------------------------------------------------

NFA_IF:		db	82h
		db	"IF"
		dw	NFA_ZGREAT
		db	42h
;
IFF:		call	DROP
		pop	R6
		pop	R7
		ld	R10, #18h
		ldci	@R10, @RR6
		ldci	@R10, @RR6
		or	R4, R5
		jr	NZ, iff1
		jp	@RR8
iff1:		jp	@RR6

;------------------------------------------------------------------------------
; ELSE ( -- )
; Dieses Wort trennt in einer Verzweigung die davor stehende Wortfolge
; fr den Ja-Zweig (wahr) von der danach stehenden fr den Nein-Zweig (falsch).
;------------------------------------------------------------------------------

NFA_ELSEE:	db	84h
		db	"ELSE"
		dw	NFA_IF
		db	52h
;
ELSEE:		pop	R6
		pop	R7
		ld	R10, #18h
		ldci	@R10, @RR6
		ldci	@R10, @RR6
		jp	@RR8

;------------------------------------------------------------------------------
; THEN (-- )
; Ende einer Verzweigung. Die folgenden Wrter werden unabhngig von der
; davor stehenden Verzweigung ausgefhrt.
;------------------------------------------------------------------------------

NFA_THENN:	db	84h
		db	"THEN"
		dw	NFA_ELSEE
		db	10h
;
THENN:		ret

;------------------------------------------------------------------------------
; /MOD (n1 n2 -- n3 n4)
; n3 ist der Quotient n1/n2 und n4 der Rest aus dieser Division.
;------------------------------------------------------------------------------

NFA_SMOD:	db	84h
		db	"/MOD"
		dw	NFA_THENN
		db	0
;
		call	DROP
		call	0E0h		; in TINY-BASIC-ROM
		ld	R8, R2
		call	TOSTACK
		call	124h		; in TINY-BASIC-ROM
		jp	SWAP

;------------------------------------------------------------------------------
; MIN (n1 n2 -- n3)
; n3 ist die kleinere der beiden Zahlen n1 und n2.
;------------------------------------------------------------------------------

NFA_MIN:	db	83h
		db	"MIN"
		dw	NFA_SMOD
		db	0
;
		call	sub_E120
		jr	MI, min2
min1:		ld	R2, R4
		ld	R3, R5
min2:		ret

;------------------------------------------------------------------------------
; MAX (n1 n2 -- n3)
; n3 ist die grere der beiden Zahlen n1 und n2.
;------------------------------------------------------------------------------

NFA_MAX:	db	83h
		db	"MAX"
		dw	NFA_MIN
		db	0
;
		call	sub_E120
		jr	MI, min1
		ret

;------------------------------------------------------------------------------
; D+ (wd1 wd2 -- wd3)
; wd3 ist die Summe von wdl und wd2.
;------------------------------------------------------------------------------

NFA_DPLUS:	db	82h
		db	"D+"
		dw	NFA_MAX
		db	0
;
		call	sub_E12C
		add	R3, R9
		adc	R2, R8
		adc	R5, R11
		adc	R4, R10
		jp	TOSTACK

;------------------------------------------------------------------------------
; +-
;------------------------------------------------------------------------------

NFA_PLMIN:	db	82h
		db	"+-"
		dw	NFA_DPLUS
		db	0
;
		call	DROP
		xor	R4, R2
		jp	MI, NEGAT
		ret

;------------------------------------------------------------------------------
; D+-
;------------------------------------------------------------------------------

NFA_DPLMIN:	db	83h
		db	"D+-"
		dw	NFA_PLMIN
		db	0
;
		call	DROP
		xor	R4, R2
		jp	MI, DNEG
		ret

;------------------------------------------------------------------------------
; DECIMAL (--)
; Speichern der Zahlenbasis 10 in BASE.
;------------------------------------------------------------------------------

NFA_DECI:	db	87h
		db	"DECIMAL"
		dw	NFA_DPLMIN
		db	0
;
		call	LIT
		dw	10
		call	BASE
		jp	STORE

;------------------------------------------------------------------------------
; HEX (--)
; Speichern der Zahlenbasis 16 in BASE.
;------------------------------------------------------------------------------

NFA_HEX:	db	83h
		db	"HEX"
		dw	NFA_DECI
		db	0
;
		call	LIT
		dw	16
		call	BASE
		jp	STORE

;------------------------------------------------------------------------------
; 2DUP (32b -- 32b 32b)
; Duplizieren von 32b auf dem Stack.
;------------------------------------------------------------------------------

NFA_TDUP:	db	84h
		db	"2DUP"
		dw	NFA_HEX
		db	0
;
		call	OVER
		jp	OVER

;------------------------------------------------------------------------------
; C@ (adr - 8b )
; 8b ist der Inhalt der Speicherzelle adr.
;------------------------------------------------------------------------------

NFA_CAT:	db	82h
		db	"C@"
		dw	NFA_TDUP
		db	0
;
		ldc	R3, @RR2
		clr	R2
		ret

;------------------------------------------------------------------------------
; C! (8b adr -- )
; 8b wird in der Speicherzelle adr gespeichert.
;------------------------------------------------------------------------------

NFA_CSTOR:	db	82h
		db	"C!"
		dw	NFA_CAT
		db	0
;
		call	DROP
		ldc	@RR4, R3
		jp	DROP

;------------------------------------------------------------------------------
; ?
;------------------------------------------------------------------------------

NFA_QUEST:	db	81h
		db	'?'
		dw	NFA_CSTOR
		db	0
;
		call	AT
		call	DOT
		ret

;------------------------------------------------------------------------------
; TOGGLE
;------------------------------------------------------------------------------

NFA_TOGGLE:	db	86h
		db	"TOGGLE"
		dw	NFA_QUEST
		db	0
;
		call	DROP
		lde	R4, @RR2
		xor	R4, R5
		lde	@RR2, R4
		jp	DROP

;------------------------------------------------------------------------------
; CMOVE (adr1 adr 2 u --)
; Transport von u Bytes ab Adresse adr1 zur Adresse adr2
; beginnend mit dem ersten Byte.
;------------------------------------------------------------------------------

NFA_CMOVE:	db	85h
		db	"CMOVE"
		dw	NFA_TOGGLE
		db	0
;
		ld	R8, R2
		ld	R9, R3
		call	DROP
		call	DROP
cmove1:		lde	R7, @RR2
		lde	@RR4, R7
		incw	RR2
		incw	RR4
		decw	RR8
		jr	NZ, cmove1
		jp	DROP

;------------------------------------------------------------------------------
; FILL (adr u 8b --)
; Laden von u Bytes ab Speicheradresse adr mit 8b.
;------------------------------------------------------------------------------

NFA_FILL:	db	84h
		db	"FILL"
		dw	NFA_CMOVE
		db	0
;
FILL:		ld	R7, R3
		call	DROP
		call	DROP
fill1:		lde	@RR2, R7
		incw	RR2
		decw	RR4
		jr	NZ, fill1
		jp	DROP

;------------------------------------------------------------------------------
; ERASE
;------------------------------------------------------------------------------

NFA_ERASE:	db	85h
		db	"ERASE"
		dw	NFA_FILL
		db	0
;
		call	LIT
		dw	0
		jp	FILL

;------------------------------------------------------------------------------
; BLANKS
;------------------------------------------------------------------------------

NFA_BLANKS:	db	86h
		db	"BLANKS"
		dw	NFA_ERASE
		db	0
;
		call	LIT
		dec	20h
		jp	FILL

;------------------------------------------------------------------------------
; COUNT (adr1 -- adr2 +n)
; +n ist das Zhlbyte der FORTH-Zeichenkette ab adr1,
; adr2 die Adresse des ersten Zeichen.
;------------------------------------------------------------------------------

NFA_COUNT:	db	85h
		db	"COUNT"
		dw	NFA_BLANKS
		db	0
;
		lde	R5, @RR2
		clr	R4
		incw	RR2
		jp	TOSTACK

;------------------------------------------------------------------------------
; .R
;------------------------------------------------------------------------------

NFA_DOTR:	db	82h
		db	".R"
		dw	NFA_COUNT
		db	0
;
		call	SWAP
		call	DUP
		call	STOD
dotr1:		call	DABS
		call	BDIGS
		call	DIGS
		call	SIGN
dotr2:		call	EDIGS
		call	ROT
		call	OVER
		call	MINUS
		call	SPACES
		jp	TYPE

;------------------------------------------------------------------------------
; D.R (d +n --)
; Ausgabe von d als vorzeichenbehaftete Zahl rechtsbndig
; in einem Feld von +n Zeichen.
;------------------------------------------------------------------------------

NFA_DDOTR:	db	83h
		db	"D.R"
		dw	NFA_DOTR
		db	0
;
		call	ROT
		call	ROT
		call	SWAP
		call	OVER
		jr	dotr1

;------------------------------------------------------------------------------
; U.R
;------------------------------------------------------------------------------

NFA_UDOTR:	db	83h
		db	"U.R"
		dw	NFA_DDOTR
		db	0
;
		call	SWAP
		call	LIT
		dw	0
		call	BDIGS
		call	DIGS
		jr	dotr2

;------------------------------------------------------------------------------
; QUIT (--)
; Return-Stack saubern, ausfhrenden Modus einschalten
; und Dialog ber aktuellen Eingabegert beginnen.
;------------------------------------------------------------------------------

NFA_QUIT:	db	84h
		db	"QUIT"
		dw	NFA_UDOTR
		db	0
;
		call	LIT
		dw	0
		call	STATE
		call	STORE
		jp	start3

;------------------------------------------------------------------------------
; BUSIN (adr -- n)
; JU+TE-FORTH: Lesen des Registers mit der Adresse adr (untere 8 Bit)
; in die unteren 8 Bit von n
;------------------------------------------------------------------------------

NFA_BUSIN:	db	85h
		db	"BUSIN"
		dw	NFA_QUIT
		db	0
;
BUSIN:		ld	R3, @R3
		clr	R2
		ret

;------------------------------------------------------------------------------
; BUSOUT (n adr -- )
; JU+TE-FORTH: Laden des Registers mit der Adresse adr (untere 8 Bit)
; mit den unteren 8 Bit von n
;------------------------------------------------------------------------------

NFA_BUSOUT:	db	86h
		db	"BUSOUT"
		dw	NFA_BUSIN
		db	0
;
		call	DROP
		ld	@R5, R3
		jp	DROP

;------------------------------------------------------------------------------
; ALLOT ( w -- )
; Reservieren von w Bytes im Wrterbuch.
;------------------------------------------------------------------------------

NFA_ALLOT:	db	85h
		db	"ALLOT"
		dw	NFA_BUSOUT
		db	0
;
		call	HERE
		call	PLUS
		jr	LINK

;------------------------------------------------------------------------------
; LINK (adr --)
; JU+TE-FORTH: Fortsetzung des Wrterbuchs an beliebiger anderer Stelle adr,
; falls der obere Anwender-RAM zu nach an die Grenze %FBFF kommt
;------------------------------------------------------------------------------

NFA_LINK:	db	84h
		db	"LINK"
		dw	NFA_ALLOT
		db	0
;
LINK:		call	DUP
		call	HERE
		call	AT
		call	SWAP
		call	STORE
		call	DP
		jp	STORE

;------------------------------------------------------------------------------
; DEPTH (-- +n)
; +n ist die Anzahl der Stapeleintrge auer +n.
;------------------------------------------------------------------------------

NFA_DEPTH:	db	85h
		db	"DEPTH"
		dw	NFA_LINK
		db	0
;
DEPTH:		ld	R5, #50h
		sub	R5, R15
		rr	R5
		ld	R4, #0
		jp	TOSTACK

;------------------------------------------------------------------------------
; ROLL (+n -- )
; Rotieren der +n + 1 obersten Stapeleintrge auer + n.
;------------------------------------------------------------------------------

NFA_ROLL:	db	84h
		db	"ROLL"
		dw	NFA_DEPTH
		db	0
;
ROLL:		ld	R4, R3
		add	R4, R4
		add	R4, R15
		ld	R5, R4
		ld	R3, @R4
		dec	R5
		inc	R4
		ld	R6, #'1'
		cp	R4, #51h
		jp	NC, error	; Ausgabe Fehler #MSG <R6>
		ld	R2, @R4
roll1:		ld	R7, @R5
		ld	@R4, R7
		dec	R4
		dec	R5
		ld	R7, @R5
		ld	@R4, R7
		dec	R4
		dec	R5
		cp	R15, R5
		jr	C, roll1
		inc	R15
		inc	R15
		ret

;------------------------------------------------------------------------------
; PICK (+n -- 16b)
; 16b ist die Kopie des (+n + 1)ten Stapeleintrags auer +n.
;------------------------------------------------------------------------------

NFA_PICK:	db	84h
		db	"PICK"
		dw	NFA_ROLL
		db	0
;
PICK:		ld	R4, R3
		add	R4, R4
		add	R4, R15
		ld	R6, #'1'
		cp	R4, #50h
		jp	NC, error	; Ausgabe Fehler #MSG <R6>
		ld	R3, @R4
		inc	R4
		ld	R2, @R4
		ret

;------------------------------------------------------------------------------
; 2ROT (32b1 32b2 32b3 -- 32b2 32b3 32b1)
; Rotieren der drei obersten doppelt genauen Stapeleintrge,
; so da der drittoberste an die Spitze gelangt.
;------------------------------------------------------------------------------

NFA_TROT:	db	84h
		db	"2ROT"
		dw	NFA_PICK
		db	0
;
		call	LIT
		dw	5
		call	ROLL
		call	LIT
		dw	5
		jp	ROLL

;------------------------------------------------------------------------------
; 2DROP (32b -- )
; Entfernen von 32b aus dem Stack.
;------------------------------------------------------------------------------

NFA_TDROP:	db	85h
		db	"2DROP"
		dw	NFA_TROT
		db	0
;
TDROP:		call	DROP
		jp	DROP

;------------------------------------------------------------------------------
; 2SWAP (32b 32b2 -- 32b2 32b1)
; Vertauschen der beiden obersten doppelt genauen Stapeleintrge.
;------------------------------------------------------------------------------

NFA_TSWAP:	db	85h
		db	"2SWAP"
		dw	NFA_TDROP
		db	0
;
TSWAP:		call	LIT
		dw	3
		call	ROLL
		call	LIT
		dw	3
		jp	ROLL

;------------------------------------------------------------------------------
; D- (wd1 wd2 -- wd3)
; wd3: Differenz wd1 minus wd2.
;------------------------------------------------------------------------------

NFA_DMINUS:	db	82h
		db	"D-"
		dw	NFA_TSWAP
		db	0
;
		call	sub_E157
		jp	TOSTACK

;------------------------------------------------------------------------------
; DU< (ud1 ud2 -- ?)
; Flag ? ist wahr, wenn ud1 < ud2.
;------------------------------------------------------------------------------

NFA_DULESS:	db	83h
		db	"DU<"
		dw	NFA_DMINUS
		db	0
;
		call	sub_E163
		jr	NC, duless2
duless1:	decw	RR2
duless2:	ret

;------------------------------------------------------------------------------
; D< (d1 d2 -- ?)
; Flag ? ist wahr, wenn d1 < d2.
;------------------------------------------------------------------------------

NFA_DLESS:	db	82h
		db	"D<"
		dw	NFA_DULESS
		db	0
;
DLESS:		call	sub_E163
		jr	LT, duless1
		ret

;------------------------------------------------------------------------------
; D= (wd1 wd2 -- ?)
; Flag ? ist wahr, wenn wd1 = wd2.
;------------------------------------------------------------------------------

NFA_DEQU:	db	82h
		db	"D="
		dw	NFA_DLESS
		db	0
;
		call	sub_E157
dequ1:		or	R2, R3
		or	R2, R4
		or	R2, R5
		clr	R2
		clr	R3
		jr	Z, duless1
		ret

;------------------------------------------------------------------------------
; D0= (wd -- ?)
; Flag ? ist wahr, wenn wd = 0.
;------------------------------------------------------------------------------

NFA_DZEQU:	db	83h
		db	"D0="
		dw	NFA_DEQU
		db	0
;
		call	DROP
		jr	dequ1

;------------------------------------------------------------------------------
; 2OVER (32b1 32b2 -- 32b1 32b2 32b1)
; Duplizieren des zweitobersten doppelt genauen Stapeleintrages.
;------------------------------------------------------------------------------

NFA_TOVER:	db	85h
		db	"2OVER"
		dw	NFA_DZEQU
		db	0
;
TOVER:		call	LIT
		dw	3
		call	PICK
		call	LIT
		dw	3
		jp	PICK

;------------------------------------------------------------------------------
; DMAX (d1 d2 -- d3)
; d3 ist die grere der beiden Zahlen d1 und d2-
;------------------------------------------------------------------------------

NFA_DMAX:	db	84h
		db	"DMAX"
		dw	NFA_TOVER
		db	0
;
		call	TOVER
		call	TOVER
		call	DLESS
		call	IFF
		dw	dmax1
		call	TSWAP
		call	THENN
dmax1:		jp	TDROP

;------------------------------------------------------------------------------
; DMIN (d1 d2 -- d3)
; d3 ist die kleinere der beiden Zahlen d1 und d2
;------------------------------------------------------------------------------

NFA_DMIN:	db	84h
		db	"DMIN"
		dw	NFA_DMAX
		db	0
;
		call	TOVER
		call	TOVER
		call	DLESS
		call	IFF
		dw	dmin1
		call	ELSEE
		dw	dmin2
dmin1:		call	TSWAP
		call	THENN
dmin2:		jp	TDROP

;------------------------------------------------------------------------------
; U< (u1 u2 -- ?)
; Flag ? ist wahr, wenn u1 < u2.
;------------------------------------------------------------------------------

NFA_USLASH:	db	82h
		db	"U<"
		dw	NFA_DMIN
		db	0
;
USLASH:		call	sub_E120
		clr	R2
		clr	R3
		jr	C, TWOAT+2	; ???
		ret

;------------------------------------------------------------------------------
; 1- (w1 -- w2)
; w2 ist die Differenz w1 minus 1.
;------------------------------------------------------------------------------

NFA_ONEMI:	db	82h
		db	"1-"
		dw	NFA_USLASH
		db	0
;
		jr	twomi1

;------------------------------------------------------------------------------
; 2- (w1 -- w2)
; w2 ist die Differenz w1 minus 2.
;------------------------------------------------------------------------------

NFA_TWOMI:	db	82h
		db	"2-"
		dw	NFA_ONEMI
		db	0
;
		decw	RR2
twomi1:		decw	RR2
		ret

;------------------------------------------------------------------------------
; 2/ (n1 -- n2)
; n2 ist der Quotient n1/2.
;------------------------------------------------------------------------------

NFA_TWOSL:	db	82h
		db	"2/"
		dw	NFA_TWOMI
		db	0
;
		sra	R2
		rrc	R3
		ret

;------------------------------------------------------------------------------
; D2/ (d1 -- d2)
; d2 ist der Quotient d1/2.
;------------------------------------------------------------------------------

NFA_DTWOSL:	db	83h
		db	"D2/"
		dw	NFA_TWOSL
		db	0
;
		call	DROP
		sra	R4
		rrc	R5
		rrc	R2
		rrc	R3
		jp	TOSTACK

;------------------------------------------------------------------------------
; NOT (16b1 -- 16b2)
; 16b2 ist die bitweise logische Negation von 16b1 (Einserkomplement).
;------------------------------------------------------------------------------

NFA_NOT:	db	83h
		db	"NOT"
		dw	NFA_DTWOSL
		db	0
;
		com	R2
		com	R3
		ret

;------------------------------------------------------------------------------
; +! (w adr -- )
; w wird zum Inhalt der Speicherzellen adr und adr+1 addiert.
;------------------------------------------------------------------------------

NFA_PLSTO:	db	82h
		db	"+!"
		dw	NFA_NOT
		db	0
;
PLSTO:		call	DROP
		lde	R6, @RR4
		incw	RR4
		lde	R7, @RR4
		add	R7, R3
		adc	R6, R2
		lde	@RR4, R7
		decw	RR4
		lde	@RR4, R6
		jp	DROP

;------------------------------------------------------------------------------
; 2@ (adr -- 32b)
; 32b ist der Inhalt der Speicherzellen adr bis adr+3
;------------------------------------------------------------------------------

NFA_TWOAT:	db	82h
		db	"2@"
		dw	NFA_PLSTO
		db	0
;
TWOAT:		call	DUP
		call	TWOP
		call	AT
		call	SWAP
		jp	AT

;------------------------------------------------------------------------------
; 2! (32b adr -- )
; 32b wird in den Speicherzellen adr bis adr+3 gespeichert.
;------------------------------------------------------------------------------

NFA_TWOSTOR:	db	82h
		db	"2!"
		dw	NFA_TWOAT
		db	0
;
TWOSTORE:	call	DUP
		call	TOR
		call	STORE
		call	RFROM
		call	TWOP
		jp	STORE

;------------------------------------------------------------------------------
; CMOVE> (adr1 adr2 u --)
; Transport von u Bytes ab Adresse adr1 zur Adresse adr2
; beginnend mit dem letzten Byte.
;------------------------------------------------------------------------------

NFA_CMOVL:	db	86h
		db	"CMOVE>"
		dw	NFA_TWOSTOR
		db	0
;
CMOVL:		ld	R8, R2
		ld	R9, R3
		call	DROP
		call	DROP
		add	R3, R9
		adc	R2, R8
		add	R5, R9
		adc	R4, R8
cmovl1:		lde	R7, @RR2
		lde	@RR4, R7
		decw	RR2
		decw	RR4
		decw	RR8
		jr	NZ, cmovl1
		jp	DROP

;------------------------------------------------------------------------------
; FORTH-83 (--)
; Anzeige der Verfgbarkeit eines FORTH-83-Systems
;------------------------------------------------------------------------------

NFA_F83:	db	88h
		db	"FORTH-83"
		dw	NFA_CMOVL
		db	0
;
		call	DOTQ
		db	51
		db	"SUBSET +DOPP.GEN.|AUTOR:HOYER|ANGEPASST:SWB 1990   "
		
		ret

;------------------------------------------------------------------------------
; UM* (u1 u2 -- ud)
; ud ist das doppelt genaue Produkt von u1 und u2,
; alle Zahlen in vorzeichenloser Deutung.
;------------------------------------------------------------------------------

NFA_UMSTAR:	db	83h
		db	"UM*"
		dw	NFA_F83
		db	0
;
UMSTAR:		jp	USTAR

;------------------------------------------------------------------------------
; UM/MOD (ud u1 -- u2 u3)
; u2 ist der Rest und u3 der Quotient aus ud/u1,
; alle Zahlen in vorzeichenloser Deutung.
;------------------------------------------------------------------------------

NFA_UMMOD:	db	86h
		db	"UM/MOD"
		dw	NFA_UMSTAR
		db	0
;
		call	MSMOD
		jp	DROP

;------------------------------------------------------------------------------
; R@ (-- 16b)
; 16b ist die Kopie des obersten Return-Stack-Eintrags.
; Der Return-Stack wird nicht verndert.
;------------------------------------------------------------------------------

NFA_RAT:	db	82h
		db	"R@"
		dw	NFA_UMMOD
		db	0
;
		jp	RR

;------------------------------------------------------------------------------
; EXIT (--)
; Abbruch der Wortausfhrung und Rcksprung
; in die rufende Wortfolge hinter die Aufrufstelle.
;------------------------------------------------------------------------------

NFA_EXIT:	db	84h
		db	"EXIT"
		dw	NFA_RAT
		db	0
;
		pop	R4
		pop	R5
		ret

;------------------------------------------------------------------------------
; INPUT ( -- d)
; JU+TE-FORTH: Eingabe einer doppelt genauen Zahl
; Abschlu der Eingabe mit Enter!
;------------------------------------------------------------------------------

NFA_INPUT:	db	85h
		db	"INPUT"
		dw	NFA_EXIT
		db	0
;
		call	QUERY
		call	TIB
		call	AT
		call	ENCL
		call	NUMB
		ret

;------------------------------------------------------------------------------
; J ( --w)
; w ist die Kopie des Index der nchstueren Schleife.
; ACHTUNG JU+TE-FORTH: Die innere Schleife mu als extra Wort definiert sein!
;------------------------------------------------------------------------------

NFA_J:		db	81h
		db	'J'
		dw	NFA_INPUT
		db	0
;
		ld	R6, sph		; Stack pointer
		ld	R7, spl		; Stack	pointer
		add	R7, #8
		lde	R4, @RR6
		incw	RR6
		lde	R5, @RR6
		jp	TOSTACK

;------------------------------------------------------------------------------
; s.a. JU+TE 01/1990, S. 70-73
; .S  ( -- )
; Anzeige des Stackinhalts
; : .S   78 SP@ DO I BUSIN I 1+ BUSIN 256 UM* DROP + CR 2 +LOOP ;
;------------------------------------------------------------------------------

NFA_SDOT:	db	82h
		db	".S"
		dw	NFA_J
		db	0
;
SDOT:		call	LIT
		dw	4Eh	; 80-2 (S0@ 2 -)
		call	SPAT
		call	DO
sdot1:		call	IDO
		call	BUSIN
		call	IDO
		call	ONEP
		call	BUSIN
		call	LIT
		dw	100h
		call	STAR
		call	PLUS
		call	CR
		call	DOT
		call	LIT
		dw	2
		call	PLOOP
		dw	sdot1
		ret

;------------------------------------------------------------------------------
; s. JU+TE 01/1990, S. 70-73
; STACK ( -- )
; : STACK   DEPTH 0> IF S. ELSE ." LEER" THEN ;
;------------------------------------------------------------------------------

NFA_STACK:	db	85h
		db	"STACK"
		dw	NFA_SDOT
		db	0
;
STACK:		call	DEPTH
		call	ZGREAT
		call	IFF
		dw	stack1
		call	SDOT
		call	ELSEE
		dw	stack2
stack1:		call	DOTQ
		db	4,"LEER"
		call	THENN
stack2:		ret

INITDP		dw	NFA_STACK

END		equ	$

		end
